/*********************************************************************

                  S y S D S  source file

                 Davide Bucci, June 2003

 *********************************************************************/

// Refer to http://www.sfu.ca/sca/Manuals/247/midi/SDSSpec.html for
// more informations on MIDI Sample Dump Standard.

// I am sorry, even if this program is command-line driven, it does not
// seems to be very portable... In particular, I had to use several 
// Windows API to drive the low level MIDI system.

// The source of this program is distributed under the GNU general
// public licence of the Free Software Foundation.


#include "stdafx.h"

#define MAXSYSEXBUF 1024  // System exclusive buffer
#define MAXSYSEXDIM 4196  // System exclusive dimension

struct wav_fmt{
  unsigned int len;         // Sample length
  unsigned int bits_per_samples;
  unsigned int sample_freq;
} fmt;


int *SampleHeap;

int VerboseLevel;

unsigned int flag_exit=0;
unsigned int SysXFlag=0;
char *SysXBuffer;

/*********************************************************************

    Function prototypes

 *********************************************************************/

void CALLBACK midiCallback(HMIDIIN handle, UINT uMsg, DWORD dwInstance, DWORD dwParam1, DWORD dwParam2);
void PrintMidiInErrorMsg(unsigned long err);
int ProcessMidiSysEx(LPBYTE midi_sysex, unsigned int length, HMIDIOUT handleOUT);
void PrintMidiOutErrorMsg(unsigned long err);
int SendMidiSysExData(HMIDIOUT handleOUT, unsigned char *data, int length);
int MidiSendSDSData(UINT iMidiIn, UINT iMidiOut);
void DisplayHelp(void);
void vprint(int Verbosing, char *str);
void ConfigureMIDI(unsigned int *midi_in_n, unsigned int *midi_out_n);

/*********************************************************************

    The main routine and other simple accessories

 *********************************************************************/

int main(int argc, char* argv[])
{
  char filename[256]; // File name buffer
  char Buf[256];      // Printing buffer
  int i,k;            // Counters
  int f_undef=1;      // Flag: filename undefined
  unsigned int midi_in_n=-2;    // -2 means undefined
  unsigned int midi_out_n=-2;
  int c, err;
  FILE *fOut;
  
  VerboseLevel=1;     // Set default verbosing level
 
  printf("*** SySDS, version 0.9 \nDavide Bucci 2003,    try option -h  for help\n");

  // Process each argument on the command line
  for(k=1; k<argc; ++k){
    if(argv[k][0]=='-'){  // An option found
      switch(argv[k][1]){
      case 'h':     // Display an help
        DisplayHelp();
        break;
      case 'c':     // Configure midi devices
        ConfigureMIDI(&midi_in_n, &midi_out_n);
        break;
      case 'i':     // Specify MidiIn device
        midi_in_n=argv[k][2]-'0';  // Convert the number
        break;
      case 'o':     // Specify MidiOut device
        midi_out_n=argv[k][2]-'0'; // Convert the number
        break;
      case 'v':     // Verbose mode
        VerboseLevel=2;  // More verbose mode
        if(argv[k][2]=='v')
          VerboseLevel=3;
        break;
      default:
        printf("Option not recognized: -%c\n", argv[k][1]);
      }
    } else {
      if(f_undef){    // Filename not found yet
        strncpy(filename, argv[k], 256);
        f_undef=0;
      } else
        printf("Filename specified more than once in the command line!\n");
    }
  }


  // Eventually, charge the configuration file
  if(midi_in_n<-1 || midi_in_n<-1){
    fOut=fopen("settings.ini", "r");
    if(fOut!=NULL){
      fscanf(fOut, "%d%d", &midi_in_n, &midi_out_n);
      fclose(fOut);
    } else {
      printf("Could not open the configuration file\n");
      // If the configuration file is not found, call the configuration section
      ConfigureMIDI(&midi_in_n, &midi_out_n);
    }
  }


  if (f_undef){
    printf("Insert WAV file name: ");
    scanf("%256s", filename);
  } else
    strncpy(filename, argv[1], 256);


  // *************** Open the WAVE file
  FILE *fIn=fopen(filename, "rb");
  
  if(fIn==NULL){
    printf("File not found: %s\n", filename);
    exit(1);
  }
  vprint(2,"File opened\n");



  err=0;
  for(i=0; i<4;++i)
    err+=(fgetc(fIn)!="RIFF"[i]);

  for(i=0; i<4; ++i)
    fgetc(fIn);
 
  for(i=0; i<4;++i)
    err+=(fgetc(fIn)!="WAVE"[i]);

  for(i=0; i<4;++i)
    err=(fgetc(fIn)!="fmt "[i]);

  if(err) {
    printf("Not a valid .WAV file\n");
    exit(1);
  }

  for(i=0; i<8; ++i)
    fgetc(fIn);
  // Get the sampling frequency
  fmt.sample_freq=fgetc(fIn);
  fmt.sample_freq|=fgetc(fIn)<<8;
  fmt.sample_freq|=fgetc(fIn)<<16;
  fmt.sample_freq|=fgetc(fIn)<<24;
  
  for(i=0; i<6; ++i)
    fgetc(fIn);

  // Get the number of bit per sample
  fmt.bits_per_samples=fgetc(fIn);
  fgetc(fIn);

  for(i=0; i<4;++i)
    fgetc(fIn);

  fmt.len=fgetc(fIn);
  fmt.len|=fgetc(fIn)<<8;
  fmt.len|=fgetc(fIn)<<16;
  fmt.len|=fgetc(fIn)<<24;
  // len now contains the data dimension in byte
  
  fmt.len/=(fmt.bits_per_samples/8);

  sprintf(Buf, "Sampling frequency: %d Hz\n", fmt.sample_freq);
  vprint(2, Buf);
  
  sprintf(Buf, "Total length: %d samples\n", fmt.len);
  vprint(2, Buf);

  sprintf(Buf, "Bits per sample: %d\n", fmt.bits_per_samples);
  vprint(2, Buf);

  // Load data
  SampleHeap=(int*)calloc(sizeof(int), fmt.len);
  if(SampleHeap==NULL){
    printf("Error allocating WAVE buffer\n");
    exit(1);
  }

  for(i=0; i<(int)fmt.len; ++i){
    

    SampleHeap[i]=fgetc(fIn);
    
    if(fmt.bits_per_samples>8){
      // Convert the sample in complement-2
      SampleHeap[i]|=(c=fgetc(fIn)<<8);
      if(c&0x80)
        SampleHeap[i]-65536;
    }

    
   
  }

  vprint(2,"File read\n");
  
  fclose(fIn);

  return MidiSendSDSData(0,0);
}

// Verboseness-applying print routine
void vprint(int Verbosing, char *str)
{
  if(Verbosing<=VerboseLevel)
    printf(str);
}


// Show a list of all available MIDI devices and allow to choose between
// them
void ConfigureMIDI(unsigned int *pmidi_in_n, unsigned int *pmidi_out_n)
{
	MIDIOUTCAPS		moc;
	MIDIINCAPS		mic;
	long			iNumDevs, i;
	FILE  *fOut;

	/* Get the number of MIDI Out devices in this computer */
	iNumDevs = midiOutGetNumDevs();
  printf("\n Configuration\n");

	/* Go through all of those devices, displaying their names */
	for (i = -1; i < iNumDevs; i++)
	{
		if (i == -1) printf("Available MIDI Output Devices\n");
		
		/* Get info about the next device */
		if (!midiOutGetDevCaps(i, &moc, sizeof(MIDIOUTCAPS)))
		{
			/* Display its Device ID and name */
			printf("%ld: %s\n", i, moc.szPname);
		}
	}

  *pmidi_out_n=-2;
  while(*pmidi_out_n<-1 && *pmidi_out_n>(unsigned int)iNumDevs){
    printf("Select desired device: ");
    scanf("%d", pmidi_out_n);
  }
  printf("\n");

	/* Get the number of MIDI In devices in this computer */
	iNumDevs = midiInGetNumDevs();

	/* Go through all of those devices, displaying their names */
	for (i = 0; i < iNumDevs; i++)
	{
		if (!i) printf("Available MIDI Input Devices\n");
		
		/* Get info about the next device */
		if (!midiInGetDevCaps(i, &mic, sizeof(MIDIINCAPS)))
		{
			/* Display its Device ID and name */
			printf("%ld: %s\n", i, mic.szPname);
		}
  }

  *pmidi_in_n=-2;
  while(*pmidi_in_n<-1 && *pmidi_in_n>(unsigned int)iNumDevs){
    printf("Select desired device: ");
    scanf("%d", pmidi_in_n);
  }
  
  fOut=fopen("settings.ini", "w");
  if(fOut!=NULL){
    fprintf(fOut, "%d\n%d", *pmidi_in_n, *pmidi_out_n);
    fclose(fOut);
    printf("Configuration settings saved in settings.ini\n");
  } else
    printf("Could not open configuration file!\n");

  printf("\n");
}

void DisplayHelp(void)
{
  printf("\nSySDS  is a program that sends standard  .WAV files to a  MIDI device which\n"
         "must adhere to the SDS standard. Refers to the documentation of your device\n"
         "to verify its conformity.\n\n"
         "The options recognised are:\n"
         "-h        display this help screen\n"
         "-v        be verbose\n"
         "-vv       be MORE verbose\n"
         "-i0       specify the use of MIDI IN  device 0 (change this number for your\n"
         "          needs)\n"
         "-o0       specify the use of MIDI OUT device 0 (change this number for your\n"
         "          needs)\n"
         "-c        configure the default MIDI IN and MIDI OUT devices\n\n"
         "The WAV files must be either at 8 or 16 bits uncompressed\n");
}

// Sends the sample read.
int MidiSendSDSData(UINT iMidiIn, UINT iMidiOut)
{
  HMIDIIN   handleIN;
  MIDIHDR   midiHdr;
  unsigned long err;
  HMIDIOUT	handleOUT;

  // Open first the MIDI Out device
  if(err = midiOutOpen(&handleOUT, iMidiOut, 0, 0, CALLBACK_NULL)){
    printf("Error opening the MIDI Out device\n");
    exit(1);
  }


  // Open the default MIDI In device 
  // The argument dwCallBackInstance is used to pass the used Midi out handle
  if(err=midiInOpen(&handleIN, iMidiIn, (DWORD)midiCallback, (DWORD)handleOUT, CALLBACK_FUNCTION)){
    printf("Error opening midi device\n");
  } else {
    SysXBuffer=(char *) calloc(sizeof(char), MAXSYSEXBUF);
    if(SysXBuffer==NULL){
      printf("Error allocating data buffer\n");
    } else {
      midiHdr.lpData=(char *)SysXBuffer;
      midiHdr.dwBufferLength = sizeof(MAXSYSEXBUF);

      midiHdr.dwFlags=0;

      // Prepare the device
      err=midiInPrepareHeader(handleIN, &midiHdr, sizeof(MIDIHDR));

      if(err){
        printf("Error configuring midi device\n");

      } else {
        err=midiInAddBuffer(handleIN, &midiHdr, sizeof(MIDIHDR));
        if(err){
          printf("Error queueing midi buffer");
        } else {
          err=midiInStart(handleIN);
          vprint(1,"Start the transfer from your keyboard\n");
          vprint(1,"Press a key to stop\n");
          
          _getch();

          SysXFlag|=0x80;
          vprint(1, "Communication stopped\n");

          // Stop recording
          midiInReset(handleIN);
        }
      }
    }
  
    if(err) PrintMidiInErrorMsg(err);
    // Close the device
    while ((err = midiInClose(handleIN)) == MIDIERR_STILLPLAYING) Sleep(0);
    
    if(err) PrintMidiInErrorMsg(err);
 
    // Unprepare the device
    midiInUnprepareHeader(handleIN, &midiHdr, sizeof(MIDIHDR));
  }
  midiOutClose(handleOUT);
  

  return 0;
}

/*********************************************************************

    The main MIDI callback routine

 *********************************************************************/

void CALLBACK midiCallback(HMIDIIN handle, UINT uMsg, DWORD dwInstance, DWORD dwParam1, DWORD dwParam2)
{
  LPMIDIHDR     lpMIDIHeader;
  LPBYTE        data;          // Pointer to System Exclusive buffer
                               // Static variables are automatically set to 0 at the first run
  static LPBYTE midi_sysex;    // System Exclusive message container
  static unsigned int pos;     // Position between successive calls

  // The dwInstance parameter is used to pass the handle of the MIDI out device
  HMIDIOUT      handleOUT=(HMIDIOUT)dwInstance;

  switch(uMsg){
  case MIM_DATA:
    // Some MIDI regular message
    break;

  case MIM_LONGDATA:
    // Check if we are closing the MIDI device
    if(SysXFlag &0x80)
      break;

    lpMIDIHeader=(LPMIDIHDR)dwParam1;
    data=(LPBYTE)(lpMIDIHeader->lpData);

    if(!SysXFlag) {
      vprint(3,"***** System Exclusive received *****\n");
      SysXFlag |= 0x01;
      pos=0;

    }
    if (*(data + (lpMIDIHeader->dwBytesRecorded - 1)) == 0xF7) {
		  // Indicate we're done handling this particular System Exclusive message 
			SysXFlag &= (~0x01);
      
		}

    if(midi_sysex==NULL)
      midi_sysex=(LPBYTE) calloc(sizeof(BYTE), MAXSYSEXDIM);
    
    // Pack the entire System Exclusive message
    if(midi_sysex!=NULL)
      while((lpMIDIHeader->dwBytesRecorded--)) 
			  midi_sysex[pos++]=*(data)++;
					

    if(!SysXFlag){
      // The midi System Exclusive event is terminated
      ProcessMidiSysEx(midi_sysex, pos, handleOUT);  
    }

    midiInAddBuffer(handle, lpMIDIHeader, sizeof(MIDIHDR));
    break;
 /* case MIM_OPEN:
  case MIM_CLOSE:
  case MIM_ERROR: */
  case MIM_MOREDATA:
    printf("WARNING: MIM_MOREDATA occourred!\n");
  }

}

/*********************************************************************

    Print the errors occourred in MIDI transmission

 *********************************************************************/

void PrintMidiOutErrorMsg(unsigned long err)
{
  #define BUFFERSIZE 256
	char	buffer[BUFFERSIZE];
	
	if (!(err = midiOutGetErrorText(err, &buffer[0], BUFFERSIZE)))
	{
		printf("%s\r\n", &buffer[0]);
	}
	else if (err == MMSYSERR_BADERRNUM)
	{
		printf("Strange error number returned!\r\n");
	}
	else
	{
		printf("Specified pointer is invalid!\r\n");
	}
}


void PrintMidiInErrorMsg(unsigned long err)
{
	char	buffer[BUFFERSIZE];
	
	if (!(err = midiInGetErrorText(err, &buffer[0], BUFFERSIZE)))
	{
		printf("%s\r\n", &buffer[0]);
	}
	else if (err == MMSYSERR_BADERRNUM)
	{
		printf("Strange error number returned!\r\n");
	}
	else if (err == MMSYSERR_INVALPARAM)
	{
		printf("Specified pointer is invalid!\r\n");
	}

	else
	{
		printf("Unable to allocate/lock memory!\r\n");
	}
}



/*********************************************************************

    Process the midi system exclusive event

 *********************************************************************/

int ProcessMidiSysEx(LPBYTE midi_sysex, unsigned int length, HMIDIOUT handleOUT)
{
  unsigned int i;           // Counter
  unsigned int sample_no;   // Sample number requested
  static int  in_execution;
  static int packet_no;
  static unsigned int sample_length;     // Sample length in words
  static unsigned int trans_flag;        // Flag: being transmitting data
  static unsigned int sample_pos;        // Position within the sample
  unsigned int value;

  char Buf[256];

  unsigned int incr;
  
  unsigned char MessageHDR[]={    // Sample dump header 
      0xF0, 0x7E, 0x00, 
      0x01, 
      0x00, 0x00,         // Sample number
      0x08,               // Sample format
      0x14, 0x31, 0x01,   // Sample period in ns
      0x7f, 0x00, 0x00,   // Sample length (in words)
      0x00, 0x00, 0x00,   // Loop start point
      0x7f, 0x00, 0x00,   // Loop end point
      0x00,               // Loop type
      0xF7};    

  unsigned char MessageCANCEL[]={ // Cancel message
      0xF0, 0x7E, 0x00,
      0x7D, 
      0x00,               // Packet number
      0xF7};
  unsigned char MessageDATA[128]={ 
      0xF0, 0x7E, 0x00,
      0x02};

  unsigned int sample_period; // Sampling period in ns
  
  // Show the SysEx bytes
  for(i=0; i<length; ++i){
    sprintf(Buf, "0x%X\t", midi_sysex[i]);
    vprint(3, Buf);
  }
  vprint(3, "\n");

      
  if(in_execution){
    vprint(3, "A MIDI IN SysEx message will be ignored!\n");
    in_execution=0;
    return 1;
  }

  in_execution=1;
  // Little checks
  if(length<3){
    vprint(3, "This message is too small\n");
    in_execution=0;
    return 1;
  }

  if(midi_sysex[0]!=0xF0) { // This message is not a system exclusive data
    vprint(3, "This message does not contain system exclusive data (it has to start with 0xF0)\n");
    in_execution=0;
    return 1;
  }

  if(midi_sysex[1]!=0x7E) { // Not a SDS message
    vprint(3, "This message does not correspond to a SDS message\n");
    in_execution=0;
    return 1;
  }
  sprintf(Buf, "Channel: %d\n", (int)midi_sysex[2]);
  vprint(3, Buf);

  switch(midi_sysex[3]){
  case 0x01:
    break;
  case 0x02:
    break;
  case 0x03:  // Dump request received
    vprint(1, "Dump request received\n");
    if(length!=7){
      vprint(2, "Wrong packet length\n");
      break;
    }

    // Packing of two successive bytes 
    sample_no=midi_sysex[4];
    sample_no+=((unsigned int)midi_sysex[5])<<7;
    
    sprintf(Buf, "Requested sample number %d\n", sample_no);
    vprint(2, Buf);
    vprint(2, "Sending a DUMP HEADER message \n");

    // Sampling period in ns
    sample_period=(unsigned int) 1e9/fmt.sample_freq;
    // WAV length in samples
    sample_length=fmt.len;
    // sample_length is a counter which downscales until it reaches 0

    sample_pos=0;
    MessageHDR[6]=fmt.bits_per_samples;
    MessageHDR[7]=sample_period&0x7F;
    MessageHDR[8]=(sample_period>>7)&0x7F;
    MessageHDR[9]=(sample_period>>14)&0x7F;
    
    MessageHDR[10]=sample_length&0x7F;
    MessageHDR[11]=(sample_length>>7)&0x7F;
    MessageHDR[12]=(sample_length>>14)&0x7F;

    MessageHDR[16]=sample_length&0x7F;
    MessageHDR[17]=(sample_length>>7)&0x7F;
    MessageHDR[18]=(sample_length>>14)&0x7F;

    for(i=0; i<sizeof(MessageHDR); ++i){
      sprintf(Buf, "0x%X\t", MessageHDR[i]);
      vprint(3, Buf);
    }

    vprint(3, "\n");
    
    if(!SendMidiSysExData(handleOUT, MessageHDR, sizeof(MessageHDR)))
      trans_flag=1;     // We start the transmission
    

    packet_no=-1;    // Begin with the first packet

    break;
  
  case 0x7C:  // WAIT
    vprint(2, "Wait message received\n");
    break;

  case 0x7E:
    vprint(2, "NACK message received\n");
    if(!trans_flag)
      break;
    vprint(2, "Sending CANCEL message\n");
    SendMidiSysExData(handleOUT, MessageCANCEL, sizeof(MessageCANCEL));
    vprint(1, "Tranmission aborted\n");
    break;

  case 0x7F:  // ACK
    if(VerboseLevel<2)
      vprint(1, ".");

    sprintf(Buf,"Aknowledging paket number %d\n",midi_sysex[4]);
    vprint(2, Buf);
    if(!trans_flag)
      break;

    if(sample_length==0){
      vprint(1,"\nTransmission ended!\n");
      vprint(2,"Sending CANCEL message\n");
      SendMidiSysExData(handleOUT, MessageCANCEL, sizeof(MessageCANCEL));
      vprint(1,"Press a key to stop transmission\n");

      flag_exit=1;   // EXIT the program

      break;
    }
    if((int)midi_sysex[4]!=packet_no&&packet_no!=-1){
      sprintf(Buf, "Problems with the ACK sequence: the ACK %d was attended!\n", packet_no);
      vprint(3, Buf);
      vprint(3, "ACK ignored\n");
      
    }
    ++packet_no;
    packet_no&=0x7F;
    // Send data at this point

    MessageDATA[4]=packet_no;

    if(fmt.bits_per_samples==8)
      incr=2;
    else if(fmt.bits_per_samples==16)
      incr=3;

    // Fill the packet with data
    for(i=5; i<125; i+=incr){
      if(sample_length==0)  // The ending zeros of the data format
        MessageDATA[i]=0;
      else{
        
        
        value=SampleHeap[sample_pos++]+32768;

        // Pack correctly data values
        if(fmt.bits_per_samples==8){
          MessageDATA[i]=(value>>1)&0x7F;
          MessageDATA[i+1]=(value<<7)&0x7F;
        }else if(fmt.bits_per_samples==16){
          MessageDATA[i]=(value>>9)&0x7F;
          MessageDATA[i+1]=(value>>2)&0x7F;
          MessageDATA[i+2]=(value<<5)&0x7F;
        }

        
        --sample_length;
      }  
    }


    // Provide the checksum data
    for(i=1; i<125; ++i)
      MessageDATA[125]^=MessageDATA[i];

    
      
    MessageDATA[126]=0xF7;

    for(i=0; i<127; ++i){
      sprintf(Buf, "0x%X\t", MessageDATA[i]);
      vprint(3, Buf);
    }

    vprint(3, "\n");
    SendMidiSysExData(handleOUT, MessageDATA, sizeof(MessageDATA));

    break;


  case 0x7D:  // CANCEL
    vprint(1,"Transmission failed!");
    vprint(2,"CANCEL message received");
    vprint(1,"\n");
    trans_flag=0;
    break;

  default:
    break;
  }
 

  in_execution=0;

  return 0;
}


// Send MIDI System Exclusive messages on the specified device
int SendMidiSysExData(HMIDIOUT handleOUT, unsigned char *data, int length)
{
  MIDIHDR   midiHdr;
  unsigned int err;

  midiHdr.lpData=(char *)data;
  midiHdr.dwBufferLength = length;

  midiHdr.dwFlags=0;

  // Prepare the device
  err=midiOutPrepareHeader(handleOUT, &midiHdr, sizeof(MIDIHDR));
  if(err){
    printf("Could not prepare the MIDI out header\n");
    PrintMidiOutErrorMsg(err);
    return 1;
  }

  err=midiOutLongMsg(handleOUT, &midiHdr, sizeof(MIDIHDR));
  if(err){
    printf("Could not send the DATA message\n");
    PrintMidiOutErrorMsg(err);
    return 1;
  }
  return 0;
}